@@ -100,15 +100,67 @@ static void diff_save_ways(way_t *way, xmlNodePtr root_node) {
 	  g_free(id);
 	  node_chain = node_chain->next;
 	}
-
 	diff_save_tags(way->tag, node_way);
       }
     }
- 
     way = way->next;
   }
 }
 
+static void diff_save_relations(relation_t *relation, xmlNodePtr root_node) {
+
+  /* store all modfied relations */
+  while(relation) {
+    if(relation->flags) { 
+      xmlNodePtr node_rel = xmlNewChild(root_node, NULL, 
+					 BAD_CAST "relation", NULL);
+
+      diff_save_state_n_id(relation->flags, node_rel, relation->id);
+      
+      if(!(relation->flags & OSM_FLAG_DELETED)) {
+	/* additional info is only required if the relation */
+	/* hasn't been deleted */
+	member_t *member = relation->member;
+	while(member) {
+	  xmlNodePtr node_member = xmlNewChild(node_rel, NULL, 
+					     BAD_CAST "member", NULL);
+
+	  char *ref = NULL;
+	  switch(member->type) {
+	  case NODE:
+	    xmlNewProp(node_member, BAD_CAST "type", BAD_CAST "node");
+	    ref = g_strdup_printf("%ld", member->node->id);
+	    break;
+	  case WAY:
+	    xmlNewProp(node_member, BAD_CAST "type", BAD_CAST "way");
+	    ref = g_strdup_printf("%ld", member->way->id);
+	    break;
+	  case RELATION:
+	    xmlNewProp(node_member, BAD_CAST "type", BAD_CAST "relation");
+	    ref = g_strdup_printf("%ld", member->relation->id);
+	    break;
+
+	  default:
+	    break;
+	  }
+
+	  g_assert(ref);
+	  xmlNewProp(node_member, BAD_CAST "ref", BAD_CAST ref);
+	  g_free(ref);
+
+	  if(member->role) 
+	    xmlNewProp(node_member, BAD_CAST "role", BAD_CAST member->role);
+
+	  member = member->next;
+	}
+	diff_save_tags(relation->tag, node_rel);
+      }
+    }
+    relation = relation->next;
+  }
+}
+
+
 /* return true if no diff needs to be saved */
 gboolean diff_is_clean(osm_t *osm) {
   gboolean clean = TRUE;
@@ -150,6 +202,7 @@ void diff_save(project_t *project, osm_t *osm) {
 
   diff_save_nodes(osm->node, root_node);
   diff_save_ways(osm->way, root_node);
+  diff_save_relations(osm->relation, root_node);
 
   xmlSaveFormatFileEnc(diff_name, doc, "UTF-8", 1);
   xmlFreeDoc(doc);
@@ -381,7 +434,7 @@ void diff_restore_way(xmlDoc *doc, xmlNodePtr node_node, osm_t *osm) {
   /* way may be an existing way, so remove nodes to */
   /* make space for new ones */
   if(way->node_chain) {
-    printf("  removing existing tags for diff tags\n");
+    printf("  removing existing nodes for diff nodes\n");
     osm_node_chain_free(way->node_chain);
     way->node_chain = NULL;
   }
@@ -411,6 +464,102 @@ void diff_restore_way(xmlDoc *doc, xmlNodePtr node_node, osm_t *osm) {
   way->tag = xml_scan_tags(doc, node_node->children, osm);
 }
 
+void diff_restore_relation(xmlDoc *doc, xmlNodePtr node_rel, osm_t *osm) {
+  printf("Restoring relation\n");
+	      
+  item_id_t id = xml_get_prop_int(node_rel, "id", ID_ILLEGAL);
+  if(id == ID_ILLEGAL) {
+    printf("  entry missing id\n");
+    return;
+  }
+
+  int state = xml_get_prop_state(node_rel, "state");
+
+  /* evaluate properties */
+  relation_t *relation = NULL;
+  switch(state) {
+  case OSM_FLAG_NEW:
+    printf("  Restoring NEW relation\n");
+    
+    relation = g_new0(relation_t, 1);
+    relation->visible = TRUE;
+    relation->flags = OSM_FLAG_NEW;
+    relation->time = xml_get_prop_int(node_rel, "time", 0);
+    if(!relation->time) relation->time = time(NULL);
+
+    /* attach to end of relation list */
+    relation_t **lrelation = &osm->relation;
+    while(*lrelation) lrelation = &(*lrelation)->next;  
+    *lrelation = relation;
+    break;
+
+  case OSM_FLAG_DELETED:
+    printf("  Restoring DELETE flag\n");
+    
+    if((relation = osm_get_relation_by_id(osm, id)))
+      relation->flags |= OSM_FLAG_DELETED;
+    else
+      printf("  WARNING: no relation with that id found\n");
+    break;
+
+  case OSM_FLAG_DIRTY:
+    printf("  Valid id (DIRTY)\n");
+    
+    if((relation = osm_get_relation_by_id(osm, id))) 
+      relation->flags |= OSM_FLAG_DIRTY;
+    else
+      printf("  WARNING: no relation with that id found\n");
+    break;
+
+  default:
+    printf("  Illegal relation entry\n");
+    return;
+  }
+
+  if(!relation) {
+    printf("  no valid relation\n");
+    return;
+  }
+  
+  /* update id from diff */
+  relation->id = id;
+    
+  /* update members */
+    
+  /* this may be an existing relation, so remove members to */
+  /* make space for new ones */
+  if(relation->member) {
+    printf("  removing existing members for diff members\n");
+    osm_members_free(relation->member);
+    relation->member = NULL;
+  }
+
+  /* scan for members */
+  member_t **member = &relation->member;
+  xmlNode *member_node = NULL;
+  for(member_node = node_rel->children; member_node; 
+      member_node = member_node->next) {
+    if(member_node->type == XML_ELEMENT_NODE) {
+      if(strcasecmp((char*)member_node->name, "member") == 0) {
+	/* attach member to member_chain */
+	*member = osm_parse_osm_relation_member(osm, doc, member_node);
+	if(*member) 
+	  member = &((*member)->next);
+      }
+    }
+  }
+  
+  /* node may be an existing node, so remove tags to */
+  /* make space for new ones */
+  if(relation->tag) {
+    printf("  removing existing tags for diff tags\n");
+    osm_tag_free(relation->tag);
+    relation->tag = NULL;
+  }
+  
+  relation->tag = xml_scan_tags(doc, node_rel->children, osm);
+}
+
 void diff_restore(appdata_t *appdata, project_t *project, osm_t *osm) {
   char *diff_name = g_strdup_printf("%s/%s.diff", project->path, project->name);
   
@@ -458,11 +607,16 @@ void diff_restore(appdata_t *appdata, project_t *project, osm_t *osm) {
 	    if(strcasecmp((char*)node_node->name, "node") == 0) 
 	      diff_restore_node(doc, node_node, osm);
 	    
-	    if(strcasecmp((char*)node_node->name, "way") == 0) 
+	    else if(strcasecmp((char*)node_node->name, "way") == 0) 
 	      diff_restore_way(doc, node_node, osm);
+
+	    else if(strcasecmp((char*)node_node->name, "relation") == 0) 
+	      diff_restore_relation(doc, node_node, osm);
+
+	    else 
+	      printf("WARNING: item %s not restored\n", node_node->name);
 	  }
 	  node_node = node_node->next;
-
 	}
       }
     }
