@@ -25,6 +25,8 @@ typedef struct iconbar_s {
   GtkWidget *trash;
   GtkWidget *node_add;
   GtkWidget *way_add;
+  GtkWidget *way_node_add;
+  GtkWidget *way_cut;
 
   GtkWidget *cancel;
   GtkWidget *ok;
