@@ -30,6 +30,6 @@ typedef struct {
 
 
 void info_dialog(appdata_t *appdata);
-void info_tags_replace(tag_context_t *context, tag_t *tag);
+void info_tags_replace(tag_context_t *context);
 
 #endif // INFO_H
