@@ -44,6 +44,29 @@ void messagef(GtkWidget *parent, char *title, const char *fmt, ...) {
   g_free(buf);
 }
 
+gboolean yes_no_f(GtkWidget *parent, char *title, const char *fmt, ...) {
+  va_list args;
+  va_start( args, fmt );
+  char *buf = g_strdup_vprintf(fmt, args);
+  va_end( args );
+
+  printf("%s: \"%s\"\n", title, buf); 
+  
+  GtkWidget *dialog = gtk_message_dialog_new(
+		     GTK_WINDOW(parent),
+                     GTK_DIALOG_DESTROY_WITH_PARENT,
+		     GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO,
+                     buf);
+
+  gtk_window_set_title(GTK_WINDOW(dialog), title);
+
+  gboolean yes = (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_YES);
+  gtk_widget_destroy(dialog);
+
+  g_free(buf);
+  return yes;
+}
+
 void errorf(GtkWidget *parent, const char *fmt, ...) {
   va_list args;
   va_start( args, fmt );
@@ -146,6 +169,14 @@ cb_menu_redraw(GtkWidget *window, gpointer data) {
 }
 #endif
 
+static void 
+cb_menu_save_changes(GtkWidget *window, gpointer data) {
+  appdata_t *appdata = (appdata_t*)data;
+
+  diff_save(appdata->project, appdata->osm);
+}
+
+
 #ifdef USE_HILDON
 static void 
 cb_menu_fullscreen(GtkWidget *widget, gpointer data) {
@@ -284,6 +315,11 @@ void menu_create(appdata_t *appdata) {
   gtk_menu_append(GTK_MENU_SHELL(submenu), item);
   g_signal_connect(item, "activate", GTK_SIGNAL_FUNC(cb_menu_upload), appdata);
 
+  item = gtk_menu_item_new_with_label( _("Save changes...") );
+  gtk_menu_append(GTK_MENU_SHELL(submenu), item);
+  g_signal_connect(item, "activate", GTK_SIGNAL_FUNC(cb_menu_save_changes), 
+		   appdata);
+
 #if 1 // testing
   item = gtk_menu_item_new_with_label( _("Redraw...") );
   gtk_menu_append(GTK_MENU_SHELL(submenu), item);
@@ -390,13 +426,15 @@ void cleanup(appdata_t *appdata) {
   appdata->program = NULL;
 #endif
 
-  printf("waiting for gtk to shut down ...");
+  printf("waiting for gtk to shut down ");
 
   /* let gtk clean up first */
-  while(gtk_events_pending())
+  while(gtk_events_pending()) {
+    putchar('.');
     gtk_main_iteration();
+  }
 
-  printf("ok\n");
+  printf(" ok\n");
 
   /* save a dirty project file */
   if(appdata->project && appdata->project->dirty)
