@@ -19,6 +19,8 @@
 
 #include "appdata.h"
 
+#include <gdk/gdkkeysyms.h>
+
 static void map_way_color(potlatch_conf_t *potlatch, way_t *way) {
   way->color = way->fill_color = MAP_COLOR_NONE;
   way->draw_flags = 0;
@@ -36,23 +38,28 @@ static void map_way_color(potlatch_conf_t *potlatch, way_t *way) {
       way->draw_flags |= OSM_DRAW_FLAGS_POLYGON;
       
       /* default fill color is light grey */
-      way->fill_color = DEF_FILL_COLOR;
+     way->fill_color = DEF_FILL_COLOR;
     }
   }
   
   while(plcol) {
     if(osm_way_has_key_or_value(way, plcol->tag)) {
-      way->color = (plcol->color << 8) | 0xff;
-      if(plcol->stroke) way->draw_flags |= OSM_DRAW_FLAGS_STROKE;
+
+      if(plcol->color != POTLATCH_COLOR_NONE) 
+	way->color = (plcol->color << 8) | 0xff;
+
+      if(plcol->stroke) 
+	way->draw_flags |= OSM_DRAW_FLAGS_STROKE;
       
-      if(plcol->fill_color != COLOR_NONE)
+      if(plcol->fill_color != POTLATCH_COLOR_NONE)
 	way->fill_color = (plcol->fill_color << 8) | FILL_TRANSP;
     }
     plcol = plcol->next;
   }
   
   /* use dark grey/no stroke/not filled for everything unknown */
-  if(way->color == MAP_COLOR_NONE) way->color = DARK_GREY;
+  if(way->color == MAP_COLOR_NONE)
+    way->color = DARK_GREY;
 }
 
 static void map_init_colors(potlatch_conf_t *potlatch, osm_t *osm) {
@@ -98,7 +105,7 @@ static void map_statusbar(map_t *map, map_item_t *map_item) {
     /* add some tags ... */
     while(tag) {
       /* we don't have much space, so ignore created_by tag */
-      if(strcasecmp(tag->key, "created_by") != 0) {
+      if(!osm_is_creator_tag(tag)) {
 	char *old = str;
 	str = g_strdup_printf("%s, %s=%s", old, tag->key, tag->value);
 	g_free(old);
@@ -116,12 +123,25 @@ static void map_outside_error(appdata_t *appdata) {
 	 _("Items must not be placed outside the working area!"));
 }
 
-static void map_item_chain_destroy(map_item_chain_t *chain) {
+static void map_item_chain_destroy(map_item_chain_t **chainP) {
+  map_item_chain_t *chain = *chainP;
   while(chain) {
     map_item_chain_t *next = chain->next;
     canvas_item_destroy(chain->map_item->item);
     chain = next;
   }
+
+  /* way until gtks event handling has actually destroyed this item */
+  printf("waiting for item destruction ");
+  while(gtk_events_pending() || *chainP) {
+    putchar('.');
+    gtk_main_iteration();
+  }
+  printf(" ok\n");
+
+  /* the callback routine connected to this item should have been */
+  /* called by now and it should have set the chain to NULL */
+  //  g_assert(!*chainP);
 }
 
 static void map_node_select(appdata_t *appdata, node_t *node) {
@@ -526,7 +546,7 @@ static map_item_t *map_item_at(map_t *map, gint x, gint y) {
     return NULL;
   }
 
-  printf("  there's an item\n");
+  printf("  there's an item (%p)\n", item);
 
   map_item_t *map_item = canvas_item_get_user_data(item);
 
@@ -631,6 +651,7 @@ static gboolean distance_above(map_t *map, gint x, gint y, gint limit) {
   return(sx*sx + sy*sy > limit*limit);
 }
 
+/* scroll with respect to two screen positions */
 static void map_do_scroll(map_t *map, gint x, gint y) {
   gint sx, sy;
 
@@ -640,6 +661,14 @@ static void map_do_scroll(map_t *map, gint x, gint y) {
   canvas_scroll_to(map->canvas, sx, sy);
 }
 
+/* scroll a certain step */
+static void map_do_scroll_step(map_t *map, gint x, gint y) {
+  gint sx, sy;
+  canvas_get_scroll_offsets(map->canvas, &sx, &sy);
+  canvas_scroll_to(map->canvas, sx+x, sy+y);
+}
+
+
 static gboolean item_is_selected_node(map_t *map, map_item_t *map_item) {
   printf("check if item is a selected node\n");
 
@@ -687,16 +716,75 @@ static gboolean item_is_selected_node(map_t *map, map_item_t *map_item) {
   return TRUE;
 }
 
+/* return true if the item given is the currenly selected way */
+/* also return false if nothing is selected or the selection is no way */
+static gboolean item_is_selected_way(map_t *map, map_item_t *map_item) {
+  printf("check if item is the selected way\n");
+
+  if(!map_item) {
+    printf("  no item requested\n");
+    return FALSE;
+  }
+
+  if(map->selected.type == MAP_TYPE_ILLEGAL) {
+    printf("  nothing is selected\n");
+    return FALSE;
+  }
+
+  /* clicked the highlight directly */
+  if(map_item->type != MAP_TYPE_WAY) {
+    printf("  didn't click way\n");
+    return FALSE;
+  }
+
+  if(map->selected.type == MAP_TYPE_WAY) {
+    printf("  selected item is a way\n");
+
+    if(map_item->way == map->selected.way) {
+      printf("  requested item is a selected way\n");
+      return TRUE;
+    }
+    printf("  but it's not the requested one\n");
+    return FALSE;
+  }
+
+  printf("  selected item is not a way\n");
+  return FALSE;
+}
+
+
 void map_highlight_refresh(appdata_t *appdata) {
   map_t *map = appdata->map;
   map_item_t old = map->selected;
 
   printf("type to refresh is %d\n", old.type);
+  if(old.type == MAP_TYPE_ILLEGAL) 
+    return;
 
   map_item_deselect(appdata);
   map_item_select(appdata, &old);
 }
 
+static void map_way_delete(appdata_t *appdata, way_t *way) {
+  printf("deleting way #%ld from map and osm\n", way->id);
+
+  /* remove it visually from the screen */
+  map_item_chain_destroy(&way->map_item_chain);
+
+  /* also remove the visible representation of all nodes (nodes with tags) */
+  node_chain_t *chain = way->node_chain;
+  while(chain) {
+    if(chain->node->map_item_chain) 
+      map_item_chain_destroy(&chain->node->map_item_chain);
+      
+    chain = chain->next;
+  }
+
+  /* and mark it "deleted" in the database */
+  osm_way_remove_from_relation(appdata->osm, way);
+  osm_way_delete(appdata->osm, way, FALSE);
+}
+
 static void map_node_move(appdata_t *appdata, map_item_t *map_item, 
 			  gint ex, gint ey) {
 
@@ -710,7 +798,266 @@ static void map_node_move(appdata_t *appdata, map_item_t *map_item,
   printf("  was at %d %d (%f %f)\n", 
 	 node->lpos.x, node->lpos.y,
 	 node->pos.lat, node->pos.lon);
+
+
+  /* check if it was dropped onto another node */
+  node_t *touchnode = map_hl_touchnode_get_node(map);
+  if(touchnode) {
+    map_hl_touchnode_clear(map);
+
+    printf("  dropped onto node #%ld\n", touchnode->id);
+
+    if(yes_no_f(GTK_WIDGET(appdata->window), _("Join nodes?"),
+		_("Do you want to join the dragged node with the one "
+		  "you dropped it on?"))) {
+
+      /* the touchnode vanishes and is replaced by the node the */
+      /* user dropped onto it */
+
+      /* TODO: move this into a function in osm.c */
+      way_t *way = appdata->osm->way;
+      while(way) {
+	node_chain_t *chain = way->node_chain;
+	while(chain) {
+	  if(chain->node == touchnode) {
+	    printf("  found node in way #%ld\n", way->id);
 	    
+	    /* replace by node */
+	    chain->node = node;
+	    
+	    /* and adjust way references of both nodes */
+	    node->ways++;
+	    touchnode->ways--;
+	    
+	    way->flags |= OSM_FLAG_DIRTY;
+	  }
+	  chain = chain->next;
+	}
+	way = way->next;
+      }
+      
+      /* replace node in relations */
+      relation_t *relation = appdata->osm->relation;
+      while(relation) {
+	member_t *member = relation->member;
+	while(member) {
+	  if(member->type == NODE && member->node == touchnode) {
+	    printf("  found node in relation #%ld\n", relation->id);
+	    
+	    /* replace by node */
+	    member->node = node;
+	    
+	    relation->flags |= OSM_FLAG_DIRTY;
+	  }
+	  member = member->next;
+	}
+	relation = relation->next;
+      }
+
+      /* transfer tags from touchnode to node */
+      tag_t **dst = &node->tag;
+      while(*dst) dst = &((*dst)->next);  /* find end of target tag list */
+      tag_t **src = &touchnode->tag;
+      gboolean conflict = FALSE;
+      while(*src) {
+	/* check if same key but with different value is present, */
+	/* ignoring the created_by tags */
+	if(!osm_is_creator_tag(*src) && 
+	   osm_tag_key_other_value_present(node->tag, *src))
+	  conflict = TRUE;
+
+	/* don't copy "created_by" and "source" tag or tags that already */
+	/* exist in identical form */
+	if(osm_is_creator_tag(*src) || 
+	   osm_tag_key_and_value_present(node->tag, *src)) {
+	  tag_t *next = (*src)->next;
+	  osm_tag_free(*src);
+	  *src = next;
+	} else {
+	  *dst = *src;
+	  *src = (*src)->next;
+	  dst = &((*dst)->next);
+	  *dst = NULL;
+	}
+      }
+
+      /* touchnode must not have any references to ways anymore */
+      g_assert(!touchnode->ways);
+
+      /* delete touchnode */
+      /* remove it visually from the screen */
+      map_item_chain_destroy(&touchnode->map_item_chain);
+
+      /* and remove it from the data structures */
+      osm_node_remove_from_relation(appdata->osm, touchnode);
+      osm_node_delete(appdata->osm, touchnode, FALSE, TRUE);
+      
+      /* and open dialog to resolve tag collisions if necessary */
+      if(conflict) {
+	messagef(GTK_WIDGET(appdata->window), "Node tag conflict", 
+		 "The resulting node contains some conflicting tags. "
+		 "Please solve these.");
+      }
+    
+      /* check whether this will also join two ways */
+      printf("  checking if node is end of way\n");
+      guint ways2join_cnt = 0;
+      way_t *ways2join[2] = { NULL, NULL };
+      way = appdata->osm->way;
+      while(way) {
+	if(osm_way_ends_with_node(way, node)) {
+	  if(ways2join_cnt < 2) 
+	    ways2join[ways2join_cnt] = way;
+
+	  printf("  way #%ld ends with this node\n", way->id);
+	  ways2join_cnt++;
+	}
+	way = way->next;
+      }
+
+      if(ways2join_cnt > 2) {
+	messagef(GTK_WIDGET(appdata->window), _("Too many ways to join"), 
+		 _("More than two ways now end in this node. Joining more "
+		   "than two ways is not yet implemented, sorry"));
+	
+      } else if(ways2join_cnt == 2) {
+	if(yes_no_f(GTK_WIDGET(appdata->window), 
+		    _("Join ways?"),
+		    _("Do you want to join the dragged way with the one "
+		      "you dropped it on?"))) {
+
+	  printf("  about to join ways #%ld and #%ld\n",
+		 ways2join[0]->id, ways2join[1]->id);
+
+	  /* way[1] gets destroyed and attached to way[0] */
+	  /* so check if way[1] is selected and exchainge ways then */
+	  /* so that way may stay selected */
+	  if((map->selected.type == MAP_TYPE_WAY) && 
+	     (map->selected.way == ways2join[1])) {
+	    printf("  swapping ways to keep selected one alive\n");
+	    way_t *tmp = ways2join[1];
+	    ways2join[1] = ways2join[0];
+	    ways2join[0] = tmp;
+	  }
+
+	  /* take all nodes from way[1] and append them to way[0] */
+	  /* check if we have to append or prepend to way[0] */
+	  gboolean revert = FALSE;
+	  if(ways2join[0]->node_chain->node == node) {
+	    /* make "prepend" to be "append" by reverting way[0] */
+	    printf("  target prepend -> revert\n");	    
+	    revert = TRUE;
+	    osm_way_revert(ways2join[0]);
+	  } 
+	  
+	  /* verify the common node is last in the target way */
+	  node_chain_t *chain = ways2join[0]->node_chain;
+	  while(chain->next) chain = chain->next;
+	  g_assert(chain->node == node);
+	  g_assert(!chain->next);
+	  
+	  /* common node must be first in the chain to attach */
+	  if(ways2join[1]->node_chain->node != node) {
+	    printf("  source revert\n");	    
+	    osm_way_revert(ways2join[1]);
+	  }
+	  
+	  /* verify the common node is first in the source way */
+	  g_assert(ways2join[1]->node_chain->node == node);
+
+	  /* finally append source chain to target */
+	  g_assert(!chain->next);
+	  chain->next = ways2join[1]->node_chain->next;
+
+	  ways2join[1]->node_chain->next = NULL;
+
+	  /* ---------- transfer tags from way[1] to way[0] ----------- */
+	  tag_t **dst = &ways2join[0]->tag;
+	  while(*dst) dst = &((*dst)->next);  /* find end of target tag list */
+	  tag_t **src = &ways2join[1]->tag;
+	  gboolean conflict = FALSE;
+	  while(*src) {
+	    /* check if same key but with different value is present, */
+	    /* ignoring the created_by tags */
+	    if(!osm_is_creator_tag(*src) && 
+	       osm_tag_key_other_value_present(ways2join[0]->tag, *src))
+	      conflict = TRUE;
+
+	    /* don't copy "created_by" and "source" tag or tags that already */
+	    /* exist in identical form */
+	    if(osm_is_creator_tag(*src) || 
+	       osm_tag_key_and_value_present(ways2join[0]->tag, *src)) {
+	      tag_t *next = (*src)->next;
+	      osm_tag_free(*src);
+	      *src = next;
+	    } else {
+	      *dst = *src;
+	      *src = (*src)->next;
+	      dst = &((*dst)->next);
+	      *dst = NULL;
+	    }
+	  }
+
+	  /* ---- transfer relation membership from way[1] to way[0] ----- */
+	  relation_chain_t *rchain = 
+	    osm_way_to_relation(appdata->osm, ways2join[1]);
+
+	  while(rchain) {
+	    relation_chain_t *next = rchain->next;
+	    printf("way[1] is part of relation #%ld\n", rchain->relation->id);
+
+	    /* make way[0] member of the same relation */
+
+	    /* walk member chain. save role of way[1] if its being found. */
+	    /* end search either at end of chain or if way[0] was foind */
+	    /* as it's already a member of that relation */
+	    member_t **member = &rchain->relation->member;
+	    char *role = NULL;
+	    while(*member && 
+		  !(((*member)->type == WAY) && 
+		    ((*member)->way == ways2join[0]))) {
+
+	      /* save role of way[1] */
+	      if(((*member)->type == WAY) && ((*member)->way == ways2join[0]))
+		role = (*member)->role;
+
+	      member = &(*member)->next;
+	    }
+
+	    if(*member) 
+	      printf("  both ways were members of this relation\n");
+	    else {
+	      printf("  adding way[0] to relation\n");
+	      *member = g_new0(member_t, 1);
+	      (*member)->type = WAY;
+	      (*member)->way = ways2join[0];
+	      if(role) (*member)->role = g_strdup(role);
+	      member = &(*member)->next;
+
+	      rchain->relation->flags |= OSM_FLAG_DIRTY;
+	    }
+
+	    g_free(rchain);
+	    rchain = next;
+	  }
+
+	  
+	  /* and open dialog to resolve tag collisions if necessary */
+	  if(conflict) {
+	    messagef(GTK_WIDGET(appdata->window), "Way tag conflict", 
+		     "The resulting way contains some conflicting tags. "
+		     "Please solve these.");
+	  }
+	  
+	  ways2join[0]->flags |= OSM_FLAG_DIRTY;
+	  map_way_delete(appdata, ways2join[1]);
+	}
+      }
+    }
+  }
+
+  /* finally update dragged nodes position */
+
   /* convert mouse position to canvas (world) position */
   gint x, y;
   canvas_window2world(map->canvas, ex, ey, &x, &y);
@@ -718,7 +1065,7 @@ static void map_node_move(appdata_t *appdata, map_item_t *map_item,
     map_outside_error(appdata);
     return;
   }
-
+  
   node->lpos.x = x;
   node->lpos.y = y;
 
@@ -737,41 +1084,24 @@ static void map_node_move(appdata_t *appdata, map_item_t *map_item,
   if(!node->ways) 
     canvas_item_set_pos(item, &node->lpos, RADIUS);
   else {
-    /* only move item of it's not the highlight. highlight will be */
+    /* only move item if it's not the highlight. highlight will be */
     /* redrawed entirely lateron */
     if(!map_item->highlight)
       canvas_item_set_pos(item, &node->lpos, S_RADIUS);
   }
 
-  /* update ways, node is part of */
+  /* visually update ways, node is part of */
   way_t *way = osm->way;
   while(way) {
     if(osm_node_in_way(way, node)) {
-      printf("node is part of way #%ld\n", way->id);
+      printf("  node is part of way #%ld, redraw!\n", way->id);
       
-      /* rebuild ways positions */
-      
-      /* allocate space for nodes */
-      guint nodes = osm_way_number_of_nodes(way);
-      if(nodes > 1) {
-	canvas_points_t *points = canvas_points_new(nodes);
-      
-	int node = 0;
-	node_chain_t *node_chain = way->node_chain;
-	while(node_chain) {
-	  canvas_point_set_pos(points, node++, &node_chain->node->lpos);
-	  node_chain = node_chain->next;
-	}
-	
-	/* do something with points */
-	map_item_chain_t *chain = way->map_item_chain;
-	while(chain) {
-	  canvas_item_set_points(chain->map_item->item, points);
-	  chain = chain->next;
-	}
-	
-	canvas_points_free(points);
-      }
+      /* remove prior version of this way */
+      map_item_chain_destroy(&way->map_item_chain);
+
+      /* draw current way */
+      map_way_color(appdata->potlatch, way);
+      map_way_draw(map, way);
     }
     
     way = way->next;
@@ -793,7 +1123,7 @@ static void map_handle_click(appdata_t *appdata, map_t *map) {
 
   /* if we aready have something selected, then de-select it */
   map_item_deselect(appdata);
-  
+
   /* select the clicked item (if there was one) */
   if(map_item.type != MAP_TYPE_ILLEGAL) {
     switch(map_item.type) {
@@ -818,11 +1148,23 @@ static void map_touchnode_update(appdata_t *appdata, gint x, gint y) {
 
   map_hl_touchnode_clear(map);
 
-#if 0
-  /* check if we are close to another node */
-  g_assert(map->pen_down.on_item);
-  g_assert(map->pen_down.on_item->type == MAP_TYPE_NODE);
-#endif
+  node_t *cur_node = NULL;
+
+  /* the "current node" which is the one we are working on and which */
+  /* should not be highlighted depends on the action */
+  switch(map->action.type) {
+
+    /* in idle mode the dragged node is not highlighted */
+  case MAP_ACTION_IDLE:
+    g_assert(map->pen_down.on_item);
+    g_assert(map->pen_down.on_item->type == MAP_TYPE_NODE);
+    cur_node = map->pen_down.on_item->node;
+    break;
+
+  default:
+    break;
+  }
+
 
   /* check if we are close to one of the other nodes */
   canvas_window2world(appdata->map->canvas, x, y, &x, &y);
@@ -830,24 +1172,21 @@ static void map_touchnode_update(appdata_t *appdata, gint x, gint y) {
   while(!map->touchnode && node) {
     /* don't highlight the dragged node itself and don't highlight */
     /* deleted ones */
-    if(((!map->pen_down.on_item ||
-	 (node != map->pen_down.on_item->node))) && 
-       (!(node->flags & OSM_FLAG_DELETED))) {
+    if((node != cur_node) && (!(node->flags & OSM_FLAG_DELETED))) {
       gint nx = x - node->lpos.x;
       gint ny = y - node->lpos.y;
-	  
+
       if(nx*nx + ny*ny < RADIUS*RADIUS)
 	map_hl_touchnode_draw(map, node);
-
     }
     node = node->next;
   }
 
-  /* during way creation also nodes o0f the new way */
+  /* during way creation also nodes of the new way */
   /* need to be searched */
   if(!map->touchnode && map->action.way) {
     node_chain_t *chain = map->action.way->node_chain;
-    while(!map->touchnode && chain) {
+    while(!map->touchnode && chain && chain->next) {
       gint nx = x - chain->node->lpos.x;
       gint ny = y - chain->node->lpos.y;
 	  
@@ -888,6 +1227,16 @@ static void map_button_press(map_t *map, gint x, gint y) {
   /* button press */
   switch(map->action.type) {
 
+  case MAP_ACTION_WAY_NODE_ADD:
+  case MAP_ACTION_WAY_CUT:
+    if(item_is_selected_way(map, map->pen_down.on_item)) {
+      gint nx, ny;
+      canvas_window2world(map->canvas, x, y, &nx, &ny);
+      if(canvas_item_get_segment(map->pen_down.on_item->item, nx, ny) >= 0)
+	map_hl_cursor_draw(map, x, y);
+    }
+    break;
+
   case MAP_ACTION_NODE_ADD:
     map_hl_cursor_draw(map, x, y);
     break;
@@ -939,10 +1288,10 @@ static void map_button_release(map_t *map, gint x, gint y) {
       map_handle_click(map->appdata, map);
       
       if((old_sel.type != MAP_TYPE_ILLEGAL) && 
+	 (old_sel.type == map->selected.type) &&
 	 (old_sel.ptr == map->selected.ptr)) {
-	printf("re-selected same item of type %d/%d, "
-	       "pushing it to the bottom\n", old_sel.type, 
-	       map->selected.type);
+	printf("re-selected same item of type %d, "
+	       "pushing it to the bottom\n", old_sel.type);
 	
 	if(!map->selected.item) {
 	  printf("  item has no visible representation to push\n");
@@ -965,7 +1314,6 @@ static void map_button_release(map_t *map, gint x, gint y) {
       else {
 	printf("released after dragging node\n");
 	map_hl_cursor_clear(map);
-	map_hl_touchnode_clear(map);
 	
 	/* now actually move the node */
 	map_node_move(map->appdata, map->pen_down.on_item, x, y);
@@ -1036,7 +1384,7 @@ static void map_button_release(map_t *map, gint x, gint y) {
       }
 
       /* remove prior version of this way */
-      map_item_chain_destroy(map->action.way->map_item_chain);
+      map_item_chain_destroy(&map->action.way->map_item_chain);
 
       /* draw current way */
       map_way_color(map->appdata->potlatch, map->action.way);
@@ -1044,6 +1392,186 @@ static void map_button_release(map_t *map, gint x, gint y) {
     }
     break;
 
+  case MAP_ACTION_WAY_NODE_ADD:
+    printf("released after WAY NODE ADD\n");
+    map_hl_cursor_clear(map);
+
+    /* check if we are still hovering above the selected way */
+    map_item_t *item = map_item_at(map, x, y);
+    if(item && item_is_selected_way(map, item)) {
+      /* convert mouse position to canvas (world) position */
+      canvas_window2world(map->canvas, x, y, &x, &y);
+      gint insert_after = canvas_item_get_segment(item->item, x, y)+1;
+      if(insert_after > 0) {
+	/* create new node */
+	node_t* node = osm_node_new(map->appdata->osm, x, y);
+	osm_node_attach(map->appdata->osm, node);
+
+	/* insert it into ways chain of nodes */
+	way_t *way = item->way;
+
+	/* search correct position */
+	node_chain_t **chain = &way->node_chain;
+	while(insert_after--) {
+	  g_assert(*chain);
+	  chain = &(*chain)->next;
+	}
+
+	/* and actually do the insertion */
+	node_chain_t *new_chain_item = g_new0(node_chain_t, 1);
+	new_chain_item->node = node;
+	new_chain_item->next = *chain;
+	*chain = new_chain_item;
+
+	/* clear selection */
+	map_item_deselect(map->appdata);
+
+	/* remove prior version of this way */
+	map_item_chain_destroy(&way->map_item_chain);
+
+	/* draw the updated way */
+	map_way_draw(map, way);
+
+	/* remember that this node is contained in one way */
+	node->ways=1;
+
+	/* and that the way needs to be uploaded */
+	way->flags |= OSM_FLAG_DIRTY;
+
+	/* put gui into idle state */
+	map_action_set(map->appdata, MAP_ACTION_IDLE);
+
+	/* and redo it */
+       	map_way_select(map->appdata, way);
+      }
+    }
+    break;
+
+  case MAP_ACTION_WAY_CUT:
+    printf("released after WAY CUT\n");
+    map_hl_cursor_clear(map);
+
+    /* check if we are still hovering above the selected way */
+    item = map_item_at(map, x, y);
+    if(item && item_is_selected_way(map, item)) {
+      /* convert mouse position to canvas (world) position */
+      canvas_window2world(map->canvas, x, y, &x, &y);
+      gint insert_after = canvas_item_get_segment(item->item, x, y);
+      if(insert_after >= 0) {
+
+	/* create a duplicate of the currently selected way */
+	way_t *way = item->way;
+	way_t *new = osm_way_new();
+    
+	/* ------------  copy all tags ------------- */
+	tag_t **dst_tag = &new->tag;
+	tag_t *src_tag = way->tag;
+	while(src_tag) {
+	  *dst_tag = g_new0(tag_t, 1);
+	  (*dst_tag)->key = g_strdup(src_tag->key);
+	  (*dst_tag)->value = g_strdup(src_tag->value);
+	  dst_tag = &(*dst_tag)->next;
+	  src_tag = src_tag->next;
+	}
+
+	/* ---- transfer relation membership from way to new ----- */
+	relation_chain_t *rchain = osm_way_to_relation(map->appdata->osm, way);
+
+	while(rchain) {
+	  relation_chain_t *next = rchain->next;
+	  printf("way is part of relation #%ld\n", rchain->relation->id);
+
+	  /* make new member of the same relation */
+
+	  /* walk member chain. save role of way if its being found. */
+	  member_t **member = &rchain->relation->member;
+	  char *role = NULL;
+	  while(*member) { 
+	    /* save role of way */
+	    if(((*member)->type == WAY) && ((*member)->way == way))
+	      role = (*member)->role;
+	    member = &(*member)->next;
+	  }
+
+	  printf("  adding new to relation\n");
+	  *member = g_new0(member_t, 1);
+	  (*member)->type = WAY;
+	  (*member)->way = new;
+	  if(role) (*member)->role = g_strdup(role);
+	  member = &(*member)->next;
+	  
+	  rchain->relation->flags |= OSM_FLAG_DIRTY;
+
+	  g_free(rchain);
+	  rchain = next;
+	}
+
+	/* move parts of node_chain to the new way */
+	node_chain_t *chain = way->node_chain;
+	while(insert_after--) {
+	  g_assert(chain);
+	  chain = chain->next;
+	}
+	 
+	/* attach remaining nodes to new way */
+	new->node_chain = chain->next;
+
+	/* terminate remainig chain on old way */
+	chain->next = NULL;
+
+	/* now move the way itself into the main data structure */
+	osm_way_attach(map->appdata->osm, new);
+	
+	/* clear selection */
+	map_item_deselect(map->appdata);
+
+	/* remove prior version of this way */
+	printf("remove visible version of way #%ld\n", way->id);
+	map_item_chain_destroy(&way->map_item_chain);
+
+	/* one of the way may only consist of a single node. */
+	/* remove it then */
+	if(osm_way_number_of_nodes(way) < 2) {
+	  printf("original way has less than 2 nodes left, deleting it\n");
+	  map_way_delete(map->appdata, way);
+	  item = NULL;
+	} else {
+	  printf("original way still has %d nodes\n",
+		 osm_way_number_of_nodes(way));
+
+	  /* draw the updated old way */
+	  map_way_color(map->appdata->potlatch, way);
+	  map_way_draw(map, way);
+
+	  /* remember that the way needs to be uploaded */
+	  way->flags |= OSM_FLAG_DIRTY;
+	}
+
+	if(osm_way_number_of_nodes(new) < 2) {
+	  printf("new way has less than 2 nodes, deleting it\n");
+	  map_way_delete(map->appdata, new);
+	  new = NULL;
+	} else {
+
+	  /* colorize the new way before drawing */
+	  map_way_color(map->appdata->potlatch, new);
+	  map_way_draw(map, new);
+	}
+
+	/* put gui into idle state */
+	map_action_set(map->appdata, MAP_ACTION_IDLE);      
+
+	/* and redo selection if way still exists */
+	if(item) 
+	  map_way_select(map->appdata, way);
+	else if(new)
+	  map_way_select(map->appdata, new);
+      }
+    }
+
+    break;
+
+
   default:
     break;
   }
@@ -1128,6 +1656,18 @@ static gboolean map_motion_notify_event(GtkWidget *widget,
     map_touchnode_update(appdata, x, y);
     break;
 
+  case MAP_ACTION_WAY_NODE_ADD:
+  case MAP_ACTION_WAY_CUT:
+    map_hl_cursor_clear(map);
+    map_item_t *item = map_item_at(map, x, y);
+    if(item && item_is_selected_way(map, item)) {
+      gint nx, ny;
+      canvas_window2world(map->canvas, x, y, &nx, &ny);
+      if(canvas_item_get_segment(item->item, nx, ny) >= 0)
+	map_hl_cursor_draw(map, x, y);
+    }
+    break;
+
   default:
     break;
   }
@@ -1144,6 +1684,33 @@ gboolean map_key_press_event(appdata_t *appdata, GdkEventKey *event) {
   if(event->type == GDK_KEY_PRESS) {
     switch(event->keyval) {
 
+    case GDK_Left:
+      map_do_scroll_step(appdata->map, -50, 0);
+      break;
+
+    case GDK_Right:
+      map_do_scroll_step(appdata->map, +50, 0);
+      break;
+
+    case GDK_Up:
+      map_do_scroll_step(appdata->map, 0, -50);
+      break;
+
+    case GDK_Down:
+      map_do_scroll_step(appdata->map, 0, +50);
+      break;
+
+#ifdef USE_HILDON
+    case HILDON_HARDKEY_SELECT:
+#else
+    case 'i':
+#endif
+      if(GTK_WIDGET_FLAGS(appdata->iconbar->info) & GTK_SENSITIVE)
+	info_dialog(appdata);
+      else
+	printf("info button not enabled\n");
+      break;
+
 #ifdef USE_HILDON
     case HILDON_HARDKEY_INCREASE:
 #else
@@ -1165,9 +1732,11 @@ gboolean map_key_press_event(appdata_t *appdata, GdkEventKey *event) {
       printf("zoom is now %f\n", appdata->map->zoom);
       return TRUE;
       break;
-    }
 
-    printf("key event %d\n", event->keyval);
+    default:
+      printf("key event %d\n", event->keyval);
+      break;
+    }
   }
 
   return FALSE;
@@ -1308,9 +1877,9 @@ void map_action_set(appdata_t *appdata, map_action_t action) {
   appdata->map->action.type = action;
 
   /* enable/disable ok/cancel buttons */
-  // MAP_ACTION_IDLE=0, _NODE_ADD, _BG_ADJUST, _WAY_ADD
-  const gboolean ok_state[] = { FALSE, FALSE, TRUE, FALSE };
-  const gboolean cancel_state[] = { FALSE, TRUE, TRUE, TRUE };
+  // MAP_ACTION_IDLE=0, NODE_ADD, BG_ADJUST, WAY_ADD, WAY_NODE_ADD, WAY_CUT
+  const gboolean ok_state[] = { FALSE, FALSE, TRUE, FALSE, FALSE, FALSE };
+  const gboolean cancel_state[] = { FALSE, TRUE, TRUE, TRUE, TRUE, TRUE };
 
   g_assert(MAP_ACTION_NUM == sizeof(ok_state)/sizeof(gboolean));
   g_assert(action < sizeof(ok_state)/sizeof(gboolean));
@@ -1347,6 +1916,8 @@ void map_action_set(appdata_t *appdata, map_action_t action) {
     _("Place a node"),
     _("Adjust background image position"),
     _("Place first node of new way"),
+    _("Place node on selected way"),
+    _("Select segment to cut way"),
   };
 
   g_assert(MAP_ACTION_NUM == sizeof(str_state)/sizeof(char*));
@@ -1383,8 +1954,7 @@ void map_action_cancel(appdata_t *appdata) {
     map->action.way->node_chain = NULL;
 
     /* remove ways visual representation */
-    map_item_chain_destroy(map->action.way->map_item_chain);
-    map->action.way->map_item_chain = NULL;
+    map_item_chain_destroy(&map->action.way->map_item_chain);
 
     osm_way_free(map->action.way);
     map->action.way = NULL;
@@ -1423,6 +1993,12 @@ void map_action_ok(appdata_t *appdata) {
       printf("    node #%ld (used by %d)\n", 
 	     chain->node->id, chain->node->ways);
 
+      /* a node may have been a stand-alone node before, so remove its */
+      /* visual representation as its now drawn as part of the way */
+      /* (if at all) */
+      if(chain->node->id != ID_ILLEGAL) 
+	map_item_chain_destroy(&chain->node->map_item_chain);
+
       /* we can be sure that no node gets inserted twice (even if twice in */
       /* the ways chain) because it gets assigned a non-ID_ILLEGAL id when */
       /* being moved to the osm node chain */
@@ -1474,7 +2050,7 @@ void map_delete_selected(appdata_t *appdata) {
     if(way_chain) {
       gboolean short_way = FALSE;
       
-      /* free the chain of relations */
+      /* free the chain of ways */
       while(way_chain && !short_way) {
 	way_chain_t *next = way_chain->next;
 
@@ -1486,49 +2062,20 @@ void map_delete_selected(appdata_t *appdata) {
       }
 
       if(short_way) {
-	GtkWidget *dialog = gtk_message_dialog_new(
-		   GTK_WINDOW(appdata->window),
-		   GTK_DIALOG_DESTROY_WITH_PARENT,
-		   GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO,
-		   _("Deleting this node will also delete one or more ways "
-		     "since they'll contain only one node afterwards. "
-		     "Do you really want this?"));
-      
-	gtk_window_set_title(GTK_WINDOW(dialog), 
-			     _("Delete node in short way(s)?"));
-  
-	/* just return here if user doesn't want this */
-	if(GTK_RESPONSE_NO == gtk_dialog_run(GTK_DIALOG(dialog))) {
-	  gtk_widget_destroy(dialog);
+	if(!yes_no_f(GTK_WIDGET(appdata->window), 
+		     _("Delete node in short way(s)?"),
+		     _("Deleting this node will also delete one or more ways "
+		       "since they'll contain only one node afterwards. "
+		       "Do you really want this?")))
 	  return;
-	}
-  
-	gtk_widget_destroy(dialog);
       }
     }
 
-    /* check if this node is part of a relation */
-    relation_chain_t *rel_chain = 
-      osm_node_to_relation(appdata->osm, item.node);
-    if(rel_chain) {
-      
-      /* free the chain of relations */
-      while(rel_chain) {
-	relation_chain_t *next = rel_chain->next;
-	g_free(rel_chain);
-	rel_chain = next;
-      }
-
-      messagef(GTK_WIDGET(appdata->window), _("Node in relation"), 
-       _("This node cannot be deleted (yet) as it's part of a relation"));
-
-      return;
-    }
-
     /* remove it visually from the screen */
-    map_item_chain_destroy(item.node->map_item_chain);
+    map_item_chain_destroy(&item.node->map_item_chain);
 
     /* and mark it "deleted" in the database */
+    osm_node_remove_from_relation(appdata->osm, item.node);
     way_chain_t *chain = osm_node_delete(appdata->osm, item.node, FALSE, TRUE);
 
     /* redraw all affected ways */
@@ -1539,12 +2086,7 @@ void map_delete_selected(appdata_t *appdata) {
 	/* this way now only contains one node and thus isn't a valid */
 	/* way anymore. So it'll also get deleted (which in turn may */
 	/* cause other nodes to be deleted as well) */
-
-	/* remove it visually from the screen */
-	map_item_chain_destroy(chain->way->map_item_chain);
-
-	/* and mark it "deleted" in the database */
-	osm_way_delete(appdata->osm, chain->way, FALSE);
+	map_way_delete(appdata, chain->way);
       } else {
 	map_item_t item;
 	item.type = MAP_TYPE_WAY;
@@ -1561,30 +2103,7 @@ void map_delete_selected(appdata_t *appdata) {
 
   case MAP_TYPE_WAY:
     printf("request to delete way #%ld\n", item.way->id);
-
-    /* check if this way is part of a relation */
-    rel_chain = osm_way_to_relation(appdata->osm, item.way);
-    if(rel_chain) {
-      
-      /* free the chain of relations */
-      while(rel_chain) {
-	relation_chain_t *next = rel_chain->next;
-	g_free(rel_chain);
-	rel_chain = next;
-      }
-
-      messagef(GTK_WIDGET(appdata->window), _("Way in relation"), 
-       _("This way cannot be deleted (yet) as it's part of a relation"));
-
-      return;
-    }
-
-    /* remove it visually from the screen */
-    map_item_chain_destroy(item.way->map_item_chain);
-
-    /* and mark it "deleted" in the database */
-    osm_way_delete(appdata->osm, item.way, FALSE);
-
+    map_way_delete(appdata, item.way);
     break;
 
   default:
