@@ -39,13 +39,20 @@ void map_hl_cursor_clear(map_t *map) {
 }
 
 /* create a new item used for touched node */
-void map_hl_touchnode_draw(map_t *map, gint x, gint y) {
+void map_hl_touchnode_draw(map_t *map, node_t *node) {
   if(map->touchnode) 
     canvas_item_destroy(map->touchnode);
 
   map->touchnode = 
-    canvas_circle_new(map, CANVAS_GROUP_DRAW, x, y, 
+    canvas_circle_new(map, CANVAS_GROUP_DRAW, node->lpos.x, node->lpos.y, 
 		      2*RADIUS, 0, TOUCHNODE_COLOR, NO_COLOR);
+
+  canvas_item_set_user_data(map->touchnode, node);
+}
+
+node_t *map_hl_touchnode_get_node(map_t *map) {
+  if(!map->touchnode) return NULL;
+  return (node_t*)canvas_item_get_user_data(map->touchnode);
 }
 
 void map_hl_touchnode_clear(map_t *map) {
