@@ -228,6 +228,53 @@ tag_t *osm_parse_osm_tag(osm_t *osm, xmlDocPtr doc, xmlNode *a_node) {
   return tag;
 }
 
+gboolean osm_is_creator_tag(tag_t *tag) {
+  if(strcasecmp(tag->key, "created_by") == 0) return TRUE;
+  if(strcasecmp(tag->key, "source") == 0) return TRUE;
+
+  return FALSE;
+}
+
+gboolean osm_tag_key_and_value_present(tag_t *haystack, tag_t *tag) {
+  while(haystack) {
+    if((strcasecmp(haystack->key, tag->key) == 0) &&
+       (strcasecmp(haystack->value, tag->value) == 0))
+      return TRUE;
+
+    haystack = haystack->next;
+  }
+  return FALSE;
+}
+
+gboolean osm_tag_key_other_value_present(tag_t *haystack, tag_t *tag) {
+  while(haystack) {
+    if((strcasecmp(haystack->key, tag->key) == 0) &&
+       (strcasecmp(haystack->value, tag->value) != 0)) 
+      return TRUE;
+    
+    haystack = haystack->next;
+  }
+  return FALSE;
+}
+
+gboolean osm_way_ends_with_node(way_t *way, node_t *node) {
+  /* and deleted way may even not contain any nodes at all */
+  /* so ignore it */
+  if(way->flags & OSM_FLAG_DELETED)
+    return FALSE;
+
+  /* any valid way must have at least two nodes */
+  g_assert(way->node_chain && way->node_chain->next);
+
+  node_chain_t *chain = way->node_chain;
+  if(chain->node == node) return TRUE;
+
+  while(chain->next) chain = chain->next;
+  if(chain->node == node) return TRUE;
+
+  return FALSE;
+}
+
 /* ------------------- node handling ------------------- */
 
 void osm_node_free(node_t *node) {
@@ -487,22 +534,27 @@ static way_t *osm_parse_osm_way(osm_t *osm,
 
 /* ------------------- relation handling ------------------- */
 
+static void osm_member_free(member_t *member) {
+  if(member->role) g_free(member->role);
+  g_free(member);
+}
+
+void osm_members_free(member_t *member) {
+  while(member) {
+    member_t *next = member->next;
+    osm_member_free(member);
+    member = next;
+  }
+}
+
 static void osm_relations_free(relation_t *relation) {
   while(relation) {
     relation_t *next = relation->next;
 
     osm_tags_free(relation->tag);
-
-    member_t *member = relation->member;
-    while(member) {
-      member_t *next = member->next;
-      if(member->role) g_free(member->role);
-      g_free(member);
-      member = next;
-    }
+    osm_members_free(relation->member);
 
     g_free(relation);
-
     relation = next;
   }
 }
@@ -559,7 +611,7 @@ void osm_relations_dump(relation_t *relation) {
   }
 }
 
-static member_t *osm_parse_osm_relation_member(osm_t *osm, 
+member_t *osm_parse_osm_relation_member(osm_t *osm, 
 			  xmlDocPtr doc, xmlNode *a_node) {
   char *prop;
   member_t *member = g_new0(member_t, 1);
@@ -963,7 +1015,7 @@ static void osm_generate_tags(tag_t *tag, xmlNodePtr node) {
   }
 }
 
-/* build xml representation for a node or way */
+/* build xml representation for a way */
 char *osm_generate_xml(osm_t *osm, type_t type, void *item) {
   char str[32];
   xmlChar *result = NULL;
@@ -1016,8 +1068,57 @@ char *osm_generate_xml(osm_t *osm, type_t type, void *item) {
     } 
     break;
 
+  case RELATION:
+    {
+      relation_t *relation = (relation_t*)item;
+      xmlNodePtr rel_node = xmlNewChild(root_node, NULL, 
+					BAD_CAST "relation", NULL);
+      snprintf(str, sizeof(str), "%u", (unsigned)relation->id);
+      xmlNewProp(rel_node, BAD_CAST "id", BAD_CAST str);
+      
+      member_t *member = relation->member;
+      while(member) {
+	xmlNodePtr m_node = xmlNewChild(rel_node,NULL,BAD_CAST "member", NULL);
+	char *str = NULL;
+
+	switch(member->type) {
+	case NODE:
+	  xmlNewProp(m_node, BAD_CAST "type", BAD_CAST "node");
+	  str = g_strdup_printf("%ld", member->node->id);
+	  break;
+
+	case WAY:
+	  xmlNewProp(m_node, BAD_CAST "type", BAD_CAST "way");
+	  str = g_strdup_printf("%ld", member->way->id);
+	  break;
+
+	case RELATION:
+	  xmlNewProp(m_node, BAD_CAST "type", BAD_CAST "relation");
+	  str = g_strdup_printf("%ld", member->relation->id);
+	  break;
+
+	default:
+	  break;
+	}
+
+	if(str) {
+	  xmlNewProp(m_node, BAD_CAST "ref", BAD_CAST str);
+	  g_free(str);
+	}
+
+	if(member->role)
+	  xmlNewProp(m_node, BAD_CAST "role", BAD_CAST member->role);
+	else
+	  xmlNewProp(m_node, BAD_CAST "role", BAD_CAST "");
+
+	member = member->next;
+      }
+      osm_generate_tags(relation->tag, rel_node);
+    }
+    break;
+
   default:
-    printf("neither NODE nor WAY\n");
+    printf("neither NODE nor WAY nor RELATION\n");
     g_assert(0);
     break;
   }
@@ -1042,6 +1143,11 @@ char *osm_generate_xml_way(osm_t *osm, way_t *way) {
   return osm_generate_xml(osm, WAY, way);
 }
 
+/* build xml representation for a relation */
+char *osm_generate_xml_relation(osm_t *osm, relation_t *relation) {
+  return osm_generate_xml(osm, RELATION, relation);
+}
+
 node_t *osm_get_node_by_id(osm_t *osm, item_id_t id) {
   node_t *node = osm->node;
   while(node) {
@@ -1050,7 +1156,6 @@ node_t *osm_get_node_by_id(osm_t *osm, item_id_t id) {
 
     node = node->next;
   }
-  
   return NULL;
 }
 
@@ -1062,6 +1167,17 @@ way_t *osm_get_way_by_id(osm_t *osm, item_id_t id) {
 
     way = way->next;
   }
+  return NULL;
+}
+
+relation_t *osm_get_relation_by_id(osm_t *osm, item_id_t id) {
+  relation_t *relation = osm->relation;
+  while(relation) {
+    if(relation->id == id)
+      return relation;
+
+    relation = relation->next;
+  }
   
   return NULL;
 }
@@ -1183,7 +1299,8 @@ way_chain_t *osm_node_delete(osm_t *osm, node_t *node, gboolean permanently,
 
   /* new nodes aren't stored on the server and are just deleted permanently */
   if(node->flags & OSM_FLAG_NEW) {
-    printf("About to delete NEW node -> force permanent delete\n");
+    printf("About to delete NEW node #%ld -> force permanent delete\n",
+	   node->id);
     permanently = TRUE;
   }
 
@@ -1371,11 +1488,63 @@ gboolean osm_position_within_bounds(osm_t *osm, gint x, gint y) {
   return TRUE;
 }
 
+/* remove the given node from all relations. used if the node is to */
+/* be deleted */
+void osm_node_remove_from_relation(osm_t *osm, node_t *node) {
+  relation_t *relation = osm->relation;
+  printf("removing node #%ld from all relations:\n", node->id);
+
+  while(relation) {
+    member_t **member = &relation->member;
+    while(*member) {
+      if(((*member)->type == NODE) &&
+	 ((*member)->node == node)) {
+
+	printf("  from relation #%ld\n", relation->id);
+	
+	member_t *cur = *member;
+	*member = (*member)->next;
+	osm_member_free(cur);
+
+	relation->flags |= OSM_FLAG_DIRTY;
+      } else
+	member = &(*member)->next;
+    }
+    relation = relation->next;
+  }
+}
+
+/* remove the given way from all relations */
+void osm_way_remove_from_relation(osm_t *osm, way_t *way) {
+  relation_t *relation = osm->relation;
+  printf("removing way #%ld from all relations:\n", way->id);
+
+  while(relation) {
+    member_t **member = &relation->member;
+    while(*member) {
+      if(((*member)->type == WAY) &&
+	 ((*member)->way == way)) {
+
+	printf("  from relation #%ld\n", relation->id);
+	
+	member_t *cur = *member;
+	*member = (*member)->next;
+	osm_member_free(cur);
+
+	relation->flags |= OSM_FLAG_DIRTY;
+      } else
+	member = &(*member)->next;
+    }
+    relation = relation->next;
+  }
+}
+
 void osm_way_delete(osm_t *osm, way_t *way, gboolean permanently) {
 
   /* new ways aren't stored on the server and are just deleted permanently */
   if(way->flags & OSM_FLAG_NEW) {
-    printf("About to delete NEW way -> force permanent delete\n");
+    printf("About to delete NEW way #%ld -> force permanent delete\n", 
+	   way->id);
     permanently = TRUE;
   }
 
@@ -1431,3 +1600,21 @@ void osm_way_delete(osm_t *osm, way_t *way, gboolean permanently) {
     g_assert(found == 1);
   }
 }
+
+void osm_way_revert(way_t *way) {
+  node_chain_t *new = NULL;
+
+  /* walk old chain first to last */
+  node_chain_t *old = way->node_chain;
+  while(old) {
+    node_chain_t *next = old->next;
+
+    /* and prepend each node to the new chain */
+    old->next = new;
+    new = old;
+
+    old = next;
+  }
+
+  way->node_chain = new;
+}
