@@ -146,6 +146,7 @@ osm_t *osm_parse(char *filename);
 gboolean osm_sanity_check(GtkWidget *parent, osm_t *osm);
 tag_t *osm_parse_osm_tag(osm_t *osm, xmlDocPtr doc, xmlNode *a_node);
 node_chain_t *osm_parse_osm_way_nd(osm_t *osm, xmlDocPtr doc, xmlNode *a_node);
+member_t *osm_parse_osm_relation_member(osm_t *osm, xmlDocPtr doc, xmlNode *a_node);
 void osm_dump(osm_t *osm);
 void osm_free(osm_t *osm);
 
@@ -167,15 +168,22 @@ void osm_node_chain_free(node_chain_t *node_chain);
 int osm_node_chain_length(node_chain_t *node_chain);
 void osm_node_free(node_t *node);
 
+void osm_members_free(member_t *member);
+
 void osm_tag_free(tag_t *tag);
 void osm_tags_free(tag_t *tag);
 char *osm_tag_get_by_key(tag_t *tag, char *key);
+gboolean osm_is_creator_tag(tag_t *tag);
+gboolean osm_tag_key_and_value_present(tag_t *haystack, tag_t *tag);
+gboolean osm_tag_key_other_value_present(tag_t *haystack, tag_t *tag);
 
 char *osm_generate_xml_node(osm_t *osm, node_t *node);
 char *osm_generate_xml_way(osm_t *osm, way_t *way);
+char *osm_generate_xml_relation(osm_t *osm, relation_t *relation);
 
 node_t *osm_get_node_by_id(osm_t *osm, item_id_t id);
 way_t *osm_get_way_by_id(osm_t *osm, item_id_t id);
+relation_t *osm_get_relation_by_id(osm_t *osm, item_id_t id);
 
 guint osm_way_number_of_nodes(way_t *way);
 relation_chain_t *osm_node_to_relation(osm_t *osm, node_t *node);
@@ -194,5 +202,11 @@ void osm_way_attach(osm_t *osm, way_t *way);
 
 gboolean osm_position_within_bounds(osm_t *osm, gint x, gint y);
 item_id_t osm_new_way_id(osm_t *osm);
+gboolean osm_way_ends_with_node(way_t *way, node_t *node);
+
+void osm_way_revert(way_t *way);
+
+void osm_node_remove_from_relation(osm_t *osm, node_t *node);
+void osm_way_remove_from_relation(osm_t *osm, way_t *way);
 
 #endif /* OSM_H */
