@@ -33,6 +33,9 @@ typedef struct settings_s {
   /* changed in osm_api.c */
   char *server, *username, *password;
 
+  /* changed in wms.c */
+  char *wms_server, *wms_path;
+
 } settings_t;
 
 settings_t *settings_load(void);
